@@ -15,30 +15,45 @@
 import json
 import os
 import requests
+import six
 from six.moves import urllib
 
 from oslo_concurrency import lockutils
-from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
 
 from os_brick import exception
 from os_brick.i18n import _
 from os_brick import initiator
 from os_brick.initiator.connectors import base
+from os_brick.privileged import scaleio as priv_scaleio
 from os_brick import utils
 
 LOG = logging.getLogger(__name__)
 DEVICE_SCAN_ATTEMPTS_DEFAULT = 3
+CONNECTOR_CONF_PATH = '/opt/emc/scaleio/openstack/connector.conf'
 synchronized = lockutils.synchronized_with_prefix('os-brick-')
 
 
+def io(_type, nr):
+    """Implementation of _IO macro from <sys/ioctl.h>."""
+
+    return ioc(0x0, _type, nr, 0)
+
+
+def ioc(direction, _type, nr, size):
+    """Implementation of _IOC macro from <sys/ioctl.h>."""
+
+    return direction | (size & 0x1fff) << 16 | ord(_type) << 8 | nr
+
+
 class ScaleIOConnector(base.BaseLinuxConnector):
     """Class implements the connector driver for ScaleIO."""
 
     OK_STATUS_CODE = 200
     VOLUME_NOT_MAPPED_ERROR = 84
     VOLUME_ALREADY_MAPPED_ERROR = 81
-    GET_GUID_CMD = ['/opt/emc/scaleio/sdc/bin/drv_cfg', '--query_guid']
+    GET_GUID_OP_CODE = io('a', 14)
+    RESCAN_VOLS_OP_CODE = io('a', 10)
 
     def __init__(self, root_helper, driver=None,
                  device_scan_attempts=initiator.DEVICE_SCAN_ATTEMPTS_DEFAULT,
@@ -62,6 +77,38 @@ class ScaleIOConnector(base.BaseLinuxConnector):
         self.iops_limit = None
         self.bandwidth_limit = None
 
+    def _get_guid(self):
+        try:
+            guid = priv_scaleio.get_guid(self.GET_GUID_OP_CODE)
+            LOG.info("Current sdc guid: %s", guid)
+            return guid
+        except (IOError, OSError, ValueError) as e:
+            msg = _("Error querying sdc guid: %s") % e
+            LOG.error(msg)
+            raise exception.BrickException(message=msg)
+
+    @staticmethod
+    def _get_connector_password(config_group):
+        LOG.info("Get ScaleIO connector password from configuration file")
+        try:
+            return priv_scaleio.get_connector_password(CONNECTOR_CONF_PATH,
+                                                       config_group)
+        except Exception as e:
+            msg = _("Error getting ScaleIO connector password from "
+                    "configuration file: %s") % e
+            LOG.error(msg)
+            raise exception.BrickException(message=msg)
+
+    def _rescan_vols(self):
+        LOG.info("ScaleIO rescan volumes")
+
+        try:
+            priv_scaleio.rescan_vols(self.RESCAN_VOLS_OP_CODE)
+        except (IOError, OSError) as e:
+            msg = _("Error querying volumes: %s") % e
+            LOG.error(msg)
+            raise exception.BrickException(message=msg)
+
     @staticmethod
     def get_connector_properties(root_helper, *args, **kwargs):
         """The ScaleIO connector properties."""
@@ -265,12 +312,16 @@ class ScaleIOConnector(base.BaseLinuxConnector):
     def get_config(self, connection_properties):
         self.local_sdc_ip = connection_properties['hostIP']
         self.volume_name = connection_properties['scaleIO_volname']
-        self.volume_id = connection_properties['scaleIO_volume_id']
+        # instances which were created before Newton release don't have
+        # 'scaleIO_volume_id' property, in such cases connector will resolve
+        # volume_id from volname
+        self.volume_id = connection_properties.get('scaleIO_volume_id')
         self.server_ip = connection_properties['serverIP']
         self.server_port = connection_properties['serverPort']
         self.server_username = connection_properties['serverUsername']
-        self.server_password = connection_properties['serverPassword']
-        self.server_token = connection_properties['serverToken']
+        self.server_password = self._get_connector_password(
+            connection_properties['config_group'],
+        )
         self.iops_limit = connection_properties['iopsLimit']
         self.bandwidth_limit = connection_properties['bandwidthLimit']
         device_info = {'type': 'block',
@@ -293,7 +344,7 @@ class ScaleIOConnector(base.BaseLinuxConnector):
                 "scaleIO Volume name: %(volume_name)s, SDC IP: %(sdc_ip)s, "
                 "REST Server IP: %(server_ip)s, "
                 "REST Server username: %(username)s, "
-                "iops limit:%(iops_limit)s, "
+                "iops limit: %(iops_limit)s, "
                 "bandwidth limit: %(bandwidth_limit)s."
             ), {
                 'volume_name': self.volume_name,
@@ -306,24 +357,7 @@ class ScaleIOConnector(base.BaseLinuxConnector):
             }
         )
 
-        LOG.info("ScaleIO sdc query guid command: %(cmd)s",
-                 {'cmd': self.GET_GUID_CMD})
-
-        try:
-            (out, err) = self._execute(*self.GET_GUID_CMD, run_as_root=True,
-                                       root_helper=self._root_helper)
-
-            LOG.info("Map volume %(cmd)s: stdout=%(out)s "
-                     "stderr=%(err)s",
-                     {'cmd': self.GET_GUID_CMD, 'out': out, 'err': err})
-
-        except putils.ProcessExecutionError as e:
-            msg = (_("Error querying sdc guid: %(err)s") % {'err': e.stderr})
-            LOG.error(msg)
-            raise exception.BrickException(message=msg)
-
-        guid = out
-        LOG.info("Current sdc guid: %(guid)s", {'guid': guid})
+        guid = self._get_guid()
         params = {'guid': guid, 'allowMultipleMappings': 'TRUE'}
         self.volume_id = self.volume_id or self._get_volume_id()
 
@@ -410,7 +444,8 @@ class ScaleIOConnector(base.BaseLinuxConnector):
 
     @utils.trace
     @lockutils.synchronized('scaleio', 'scaleio-')
-    def disconnect_volume(self, connection_properties, device_info):
+    def disconnect_volume(self, connection_properties, device_info,
+                          force=False, ignore_errors=False):
         """Disconnect the ScaleIO volume.
 
         :param connection_properties: The dictionary that describes all
@@ -418,6 +453,10 @@ class ScaleIOConnector(base.BaseLinuxConnector):
         :type connection_properties: dict
         :param device_info: historical difference, but same as connection_props
         :type device_info: dict
+        :type force: bool
+        :param ignore_errors: When force is True, this will decide whether to
+                              ignore errors or raise an exception once finished
+                              the operation.  Default is False.
         """
         self.get_config(connection_properties)
         self.volume_id = self.volume_id or self._get_volume_id()
@@ -432,25 +471,7 @@ class ScaleIOConnector(base.BaseLinuxConnector):
              'server_ip': self.server_ip}
         )
 
-        LOG.info("ScaleIO sdc query guid command: %(cmd)s",
-                 {'cmd': self.GET_GUID_CMD})
-
-        try:
-            (out, err) = self._execute(*self.GET_GUID_CMD, run_as_root=True,
-                                       root_helper=self._root_helper)
-            LOG.info(
-                "Unmap volume %(cmd)s: stdout=%(out)s stderr=%(err)s",
-                {'cmd': self.GET_GUID_CMD, 'out': out, 'err': err}
-            )
-
-        except putils.ProcessExecutionError as e:
-            msg = _("Error querying sdc guid: %(err)s") % {'err': e.stderr}
-            LOG.error(msg)
-            raise exception.BrickException(message=msg)
-
-        guid = out
-        LOG.info("Current sdc guid: %(guid)s", {'guid': guid})
-
+        guid = self._get_guid()
         params = {'guid': guid}
         headers = {'content-type': 'application/json'}
         request = (
@@ -487,5 +508,31 @@ class ScaleIOConnector(base.BaseLinuxConnector):
                 raise exception.BrickException(message=msg)
 
     def extend_volume(self, connection_properties):
-        # TODO(walter-boring): is this possible?
-        raise NotImplementedError
+        """Update the local kernel's size information.
+
+        Try and update the local kernel's size information
+        for a ScaleIO volume.
+        """
+
+        self._rescan_vols()
+        volume_paths = self.get_volume_paths(connection_properties)
+        if volume_paths:
+            return self.get_device_size(volume_paths[0])
+
+        # if we got here, the volume is not mapped
+        msg = (_("Error extending ScaleIO volume"))
+        LOG.error(msg)
+        raise exception.BrickException(message=msg)
+
+    def get_device_size(self, device):
+        """Get the size in bytes of a volume."""
+        (out, _err) = self._execute('blockdev', '--getsize64',
+                                    device, run_as_root=True,
+                                    root_helper=self._root_helper)
+        var = six.text_type(out.strip())
+        LOG.debug("Device %(dev)s size: %(var)s",
+                  {'dev': device, 'var': var})
+        if var.isnumeric():
+            return int(var)
+        else:
+            return None
